from collections import UserDict
from collections.abc import Mapping
import sys
from typing import Any as Any, Dict as Dict, Union as Union

from typing_extensions import TypeAlias  # noqa: F401  # used in type hints

from dirlay.optional import pathlib


Path = pathlib.Path  # type: TypeAlias
PathType = Union[Path, str]  # type: TypeAlias

# dicts

StrDict = Dict[str, Any]  # type: TypeAlias

if sys.version_info < (3, 9):  # pragma: no cover
    AnyDict = Union[StrDict, UserDict]  # type: TypeAlias
else:
    AnyDict = Union[StrDict, UserDict[str, Any]]  # type: TypeAlias

# node tree

if sys.version_info < (3, 9):  # pragma: no cover
    DictTree = Mapping  # type: TypeAlias
else:
    DictTree = Mapping[str, 'DictNode']  # type: TypeAlias
    """
    TypeAlias: User representation of directory structure to be provided as input
    to `~dirlay.Dir` constructor or `~dirlay.Dir.update` method and operations:

    >>> tree = Dir({'a': {'b.md': 'b file content'}})
    >>> tree |= {'c.md': 'c file content'}
    """

DictNode = Union[str, DictTree]  # type: TypeAlias
"""
TypeAlias: User representation of directory node — a file or a directory.
"""



try:
    from collections import UserDict
except ImportError:  # pragma: no cover

    class UserDict(object):
        pass


class NestedDict(UserDict):
    """
    General purpose dict that allows accessing nested items by delimited keys.
    """

    dict_class = dict
    sep = '/'

    def __init__(self, dict=None, **kwargs):
        self.data = self.dict_class()
        self._len = 0
        self.update(dict, **kwargs)

    def __eq__(self, other):
        try:
            return self.data == self._operand(other)
        except TypeError:
            return False

    def __len__(self):
        return self._len

    def __getitem__(self, key):
        parent, lastpart = self._traverse(key, create_parents=False, base=self.data)
        return parent[lastpart]

    def __setitem__(self, key, item):
        self._update({key: item}, base=self.data)

    def __delitem__(self, key):
        parent, lastpart = self._traverse(key, create_parents=False, base=self.data)
        self._len -= self._count(parent[lastpart])
        del parent[lastpart]

    def _count(self, item):
        if not isinstance(item, self.dict_class):
            return 1
        else:
            return 1 + sum(self._count(v) for v in item.values())

    def traverse(self, key):
        return self._traverse(key, create_parents=False, base=self.data)

    def _traverse(self, key, create_parents, base):
        parent = base
        prev, last = None, -1
        while last < len(key):
            # find next delimited part
            prev = last + 1
            last = key.find(self.sep, prev)
            if last == -1:
                return parent, key[prev:]
            part = key[prev:last]
            # traverse nested
            if part not in parent:
                if create_parents:
                    parent[part] = self.dict_class()
                    self._len += 1
                else:
                    raise KeyError(key[:last])
            parent = parent[part]
            if not isinstance(parent, dict):
                raise ValueError('Not a dictionary: {}'.format(key[:last]))

    def update(self, other=None, **kwargs):  # type: ignore[override]
        if other:
            self._update(self._operand(other), base=self.data)
        self._update(kwargs, base=self.data)

    def _update(self, other, base):
        for k in other.keys():
            v = other[k]
            parent, key = self._traverse(k, create_parents=True, base=base)
            if key not in parent:
                self._len += 1
            if isinstance(v, dict) or isinstance(v, self.dict_class):
                if key not in parent or not isinstance(parent[key], self.dict_class):
                    parent[key] = self.dict_class()
                self._update(v, base=parent[key])
            else:
                parent[key] = v

    def __iter__(self):
        for k, _, _ in self._walk(self.data, prefix=None):
            yield k

    def items(self):
        return ((k, v) for (k, v, _) in self._walk(self.data, prefix=None))

    def keys(self):
        return tuple(k for (k, _, _) in self._walk(self.data, prefix=None))

    def values(self):
        return tuple(v for (_, v, _) in self._walk(self.data, prefix=None))

    def _walk(self, entries, prefix=None):
        for key, item in entries.items():
            nested_key = key if prefix is None else self.sep.join((prefix, key))
            yield nested_key, item, entries
            if isinstance(item, self.dict_class):
                for kvp in self._walk(item, prefix=nested_key):  # Python 2 support
                    yield kvp

    def __contains__(self, key):
        if not isinstance(key, str):
            return False
        try:
            parent, name = self._traverse(key, create_parents=False, base=self.data)
            return name in parent
        except (KeyError, ValueError):
            return False

    def get(self, key, default=None):
        try:
            parent, lastpart = self._traverse(key, create_parents=False, base=self.data)
        except (KeyError, ValueError):
            return default
        else:
            return parent.get(lastpart, default)

    def __repr__(self):
        return repr(self.data)

    def __or__(self, other):
        ret = self.copy()
        ret.update(self._operand(other))
        return ret

    def __ror__(self, other):
        ret = self.__class__(self._operand(other))
        ret.update(self)
        return ret

    def __ior__(self, other):
        self.update(self._operand(other))
        return self

    @staticmethod
    def _operand(other):
        if isinstance(other, UserDict):
            return other.data
        elif isinstance(other, dict):
            return other
        else:
            raise TypeError('Not a dictionary type: {}'.format(type(other)))

    def __copy__(self):
        return self.__class__(self.data)

    def clear(self):
        self.data.clear()

    def copy(self):
        return self.__copy__()

    @classmethod
    def fromkeys(cls, iterable, value=None):
        d = cls()
        for key in iterable:
            d[key] = value
        return d



from unittest import TestCase

from dirlay import Dir

try:
    from typing import List, Tuple  # noqa: F401  # used in type hints
    from dirlay.types import DictNode, StrDict  # noqa: F401  # used in type hints
except ImportError:
    pass


class TestConstruct(TestCase):
    def test_no_args(self):  # type: () -> None
        self.assertEqual({}, Dir().data)

    def test_construct(self):  # type: () -> None
        for src, data in (
            ({'a': 'A', 'b': {}}, {'a': 'A', 'b': {}}),
            ({'a/b/c/d.txt': 'D'}, {'a': {'b': {'c': {'d.txt': 'D'}}}}),
            ({'a': {'b': {'c': {'d.txt': 'D'}}}}, {'a': {'b': {'c': {'d.txt': 'D'}}}}),
            ({'a/b/c': {'d.txt': 'D'}}, {'a': {'b': {'c': {'d.txt': 'D'}}}}),
            ({'a/b/c.txt': 'C', 'a/bb.txt': 'BB'}, {'a': {'b': {'c.txt': 'C'}, 'bb.txt': 'BB'}}),
            ({'a/b': {}, 'a/b/c.txt': 'c'}, {'a': {'b': {'c.txt': 'c'}}}),
            ({'a/b/c.txt': 'c', 'a/b': {}}, {'a': {'b': {'c.txt': 'c'}}}),
            ({'b': {}, 'a': {}}, {'a': {}, 'b': {}}),
        ):  # fmt: skip
            self.assertEqual(data, Dir(src).data)  # type: ignore
            self.assertEqual(data, Dir(Dir(src).data).data)  # type: ignore


class TestModify(TestCase):
    def test_update_other(self):  # type: () -> None
        def assertPassing(src, upd, expected):  # type: (StrDict, StrDict, StrDict) -> None
            result = Dir(src)
            result |= upd
            self.assertEqual(Dir(expected), result)
            self.assertEqual(Dir(expected), Dir(src) | upd)
            self.assertEqual(Dir(expected), Dir(src) | Dir(upd).data)

        for src, upd, expected in (
            ({'a/b.md': 'B'}, {'a/d.md': 'D', 'e.md': 'E'}, {'a': {'b.md': 'B', 'd.md': 'D'}, 'e.md': 'E'}),
            ({'a/b.md': 'B'}, {'a/b.md': 'UPD'}, {'a/b.md': 'UPD'}),
            ({'a/b.md': 'B'}, {'a/b.md': {}}, {'a/b.md': {}}),
        ):  # fmt: skip
            assertPassing(src, upd, expected)

    def test_update_other_error(self):  # type: () -> None
        src = {'a/b.md': 'B'}
        upd = {'a/b.md/c.md': 'C'}
        with self.assertRaises(ValueError):
            Dir(src) | upd

    def test_update_data(self):  # type: () -> None
        def assertPassing(src, kvs, expected):  # type: (StrDict, List[Tuple[str, DictNode]], StrDict) -> None
            updated = Dir(src)
            for k, v in kvs:
                updated[k].data = v
                self.assertEqual(v, updated[k].data)
                self.assertEqual(Dir(expected), updated)

        for src, kvs, expected in (
            ({'a/b.md': 'B'}, [('a/b.md', 'UPD')], {'a/b.md': 'UPD'}),
        ):  # fmt: skip
            assertPassing(src, kvs, expected)  # type: ignore

    def test_update_data_error(self):  # type: () -> None
        def assertError(src, k, v, errtype):  # type: (StrDict, str, DictNode, type[Exception]) -> None
            with self.assertRaises(errtype):
                Dir(src)[k].data = v

        for src, k, v, errtype in (
            ({'a/b.md': 'B'}, 'a/b.md/c.md', 'C', ValueError),
            ({'a/b': {}}, 'a/b/c.md', 'C', KeyError),
        ):
            assertError(src, k, v, errtype)



try:
    import pathlib
except ImportError:  # pragma: no cover
    import pathlib2 as pathlib  # type: ignore

try:
    import rich
except ImportError:  # pragma: no cover
    rich = None  # type: ignore

__all__ = [
    'pathlib',
    'rich',
]



import os
from unittest import TestCase
from uuid import uuid4

from dirlay import Dir, Path, getcwd


class TestFilesystem(TestCase):
    def assertFilesystem(self, layout):  # type: (Dir) -> None
        assert layout.basedir is not None
        for node in layout.leaves():
            self.assertTrue(node.path.exists())
            if not node.is_dir:
                self.assertEqual(node.data, node.path.read_text())

    # test mktree/rmtree

    def test_error_remove_not_instantiated(self):  # type: () -> None
        with self.assertRaises(RuntimeError):
            Dir({'c.md': 'c'}).rmtree()

    def test_create_remove_tempdir(self):  # type: () -> None
        layout = Dir({'a': {'b': {'c.md': 'C'}}})
        # instantiate in temporary directory
        layout.mktree()
        assert layout.basedir is not None
        self.assertFilesystem(layout)
        # remove
        basedir = layout.basedir
        layout.rmtree()
        self.assertFalse(basedir.exists())

    def test_create_remove_userdir(self):  # type: () -> None
        layout = Dir({'a': {'b': {'c.md': 'C'}}})
        # instantiate in directory provided by user
        os.chdir('/tmp')
        layout.mktree(uuid4().hex)
        assert layout.basedir is not None
        self.assertFilesystem(layout)
        # remove
        basedir = layout.basedir
        layout.rmtree()
        self.assertFalse(basedir.exists())

    # test chdir

    def test_error_chdir_not_instantiated(self):  # type: () -> None
        with self.assertRaises(RuntimeError):
            Dir({'c.md': 'C'}).chdir('.')

    def test_error_chdir_absolute(self):  # type: () -> None
        layout = Dir({'c.md': 'C'})
        with self.assertRaises(RuntimeError):
            layout.chdir(getcwd())

    def test_chdir(self):  # type: () -> None
        layout = Dir({'a': {'b': {'c.md': 'C'}}})
        layout.mktree()
        assert layout.basedir is not None
        cwd = getcwd()
        layout.chdir('.')
        self.assertEqual(layout.basedir, getcwd())
        layout.chdir('a')
        self.assertEqual(layout.basedir / 'a', getcwd())
        layout.chdir('a/b')
        self.assertEqual(layout.basedir / 'a/b', getcwd())
        layout.rmtree()  # original cwd is restored on rmtree()
        self.assertEqual(cwd, getcwd())

    def test_mktree_chdir(self):  # type: () -> None
        cwd = getcwd()

        # default
        with Dir().mktree() as tree:
            self.assertEqual(cwd, getcwd())  # not changed

        # None
        with Dir().mktree(chdir=None) as tree:
            self.assertEqual(cwd, getcwd())  # not changed

        # False
        with Dir().mktree(chdir=False):
            self.assertEqual(cwd, getcwd())  # not changed

        # True
        with Dir().mktree(chdir=True) as tree:
            self.assertEqual(tree.basedir, getcwd())  # changed to basedir

        # root: str
        with Dir().mktree(chdir='.') as tree:
            self.assertEqual(tree.basedir, getcwd())  # changed to basedir

        # root: Path
        with Dir().mktree(chdir=Path()) as tree:
            self.assertEqual(tree.basedir, getcwd())  # changed to basedir

        # subdir: str
        with Dir({'a': {}}).mktree(chdir='a') as tree:
            self.assertEqual(tree.basedir / 'a', getcwd())  # type: ignore[operator]

        # subdir: Path
        with Dir({'a': {}}).mktree(chdir=Path('a')) as tree:
            self.assertEqual(tree.basedir / 'a', getcwd())  # type: ignore[operator]

        # subsubdir: str
        with Dir({'a': {'b': {}}}).mktree(chdir='a/b') as tree:
            self.assertEqual(tree.basedir / 'a/b', getcwd())  # type: ignore[operator]

        # subsubdir: Path
        with Dir({'a': {'b': {}}}).mktree(chdir=Path('a/b')) as tree:
            self.assertEqual(tree.basedir / 'a/b', getcwd())  # type: ignore[operator]

        # errors

        # not found
        with self.assertRaises(OSError):
            with Dir() as tree:
                tree.mktree(chdir='x')

        # invalid type
        with self.assertRaises(TypeError):
            with Dir() as tree:
                tree.mktree(chdir=[])  # type: ignore

        # absolute: str
        with self.assertRaises(ValueError):
            with Dir() as tree:
                tree.mktree(chdir='/tmp')

        # absolute: Path
        with self.assertRaises(ValueError):
            with Dir() as tree:
                tree.mktree(chdir=Path('/tmp'))



from typing import Any

try:
    from rich.tree import Tree  # type: ignore[import-not-found,unused-ignore]
except ImportError:
    Tree = None  # type: ignore[assignment,misc]  # assign to type

from dirlay import Dir

def as_tree(
    layout: Dir,
    real_basedir: bool = ...,
    show_data: bool = ...,
    **kwargs: Any,
) -> Tree: ...



import os
import shutil
import sys
from tempfile import mkdtemp

try:
    from reprlib import aRepr

    a_repr = aRepr.repr
except ImportError:  # pragma: no cover
    a_repr = repr

from dirlay.__version__ import __version__ as __version__
from dirlay.nested_dict import NestedDict as BaseNestedDict
from dirlay.optional import pathlib, rich

if sys.version_info > (3,):
    NestedDict = BaseNestedDict
else:  # pragma: no cover
    from collections import OrderedDict as BaseOrderedDict

    class OrderedDict(BaseOrderedDict):
        def __eq__(self, other):
            return dict(self) == dict(other)

        def __repr__(self):
            return '{{{}}}'.format(
                ', '.join('{!r}: {!r}'.format(k, v) for k, v in self.items())
            )

    class NestedDict(BaseNestedDict):
        dict_class = OrderedDict


if rich is not None:
    from dirlay.format_rich import as_tree, rich_print
else:  # pragma: no cover
    as_tree = None

    def rich_print(*args):
        pass


__all__ = [
    'Dir',
    'NestedDict',
    'Node',
    'Path',
    'getcwd',
]

Path = pathlib.Path


class Node(object):
    """
    Node proxy object representing directory or file.

    Attributes:

        key (``str``):
            String path relative to `~dirlay.Dir` root, and also a mapping key:

            >>> tree = Dir({'a': {'b.md': 'B'}})
            >>> tree[tree['a/b.md'].key] == tree['a/b.md']
            True

        path (`~pathlib.Path`):
            Path of the node; if parent `Dir` is not created on the file system,
            it is relative to parent `~dirlay.Dir.basedir` and matches
            `~dirlay.Node.key`; otherwise it is an absolute path on the file system.

        is_dir (``bool``):
            Whether the node is a directory.

        data (``str | dict[str, str | dict]``)
            In-memory file content if the node is a file, or, if `~dirlay.Node.is_dir`
            is ``True``, a dictionary, representing directory structure.
    """

    def __init__(self, key, path, base):
        self.key = key
        self.path = Path(path)
        self._base = base
        self._name = '.' if key == '.' else self.path.name

    def __eq__(self, other):
        return (
            isinstance(other, Node)
            and self.key == other.key
            and self.path == other.path
            and self._base is other._base
        )

    @property
    def data(self):
        return self._base[self._name]

    @data.setter
    def data(self, value):
        self._base[self._name] = value

    @property
    def is_dir(self):
        return isinstance(self.data, NestedDict.dict_class)

    def __repr__(self):
        return '<Node {!r}: {}>'.format(str(self.key), a_repr(self.data))


class Dir:
    """
    Directory layout class. See :ref:`Use cases` for examples.
    """

    def __init__(self, entries=None):
        r"""
        Example:

            >>> from dirlay import Dir

            >>> Dir({'docs/index.rst': '', 'src': {}, 'pyproject.toml': '\n'}).data
            {'docs': {'index.rst': ''}, 'src': {}, 'pyproject.toml': '\n'}

            >>> Dir({'a/b/c/d/e/f.txt': '', 'a/b/c/d/ee': {}}).data
            {'a': {'b': {'c': {'d': {'e': {'f.txt': ''}, 'ee': {}}}}}}
        """  # fmt: skip

        self._tree = NestedDict()
        if entries is not None:
            self.update(entries)
        self._basedir = None
        self._basedir_remove = False
        self._original_cwd = None

    def __repr__(self):
        return '<Dir {!r}: {}>'.format(
            str(self._basedir or '.'),
            a_repr(self._tree.data),
        )

    @property
    def data(self):
        """
        Internal data mapping.
        """
        return self._tree.data

    def __contains__(self, path):
        """
        Check whether directory layout object contains path defined.
        """
        return norm(path) in self._tree

    def __eq__(self, other):
        """
        Two directory layouts are equal if they have:

        - equal files and directories (both path and data)
        - equal `~dirlay.Dir.basedir`
        """
        return (
            isinstance(other, Dir)
            and self.basedir == other.basedir
            and self._tree == other._tree
        )

    def __getitem__(self, path):
        """
        Return `~dirlay.Node` object from string path.
        """
        key = norm(path)
        if os.path.isabs(key):
            raise ValueError('Absolute path not allowed: {!r}'.format(path))
        base, name = self._tree.traverse(key)
        if name not in base:
            raise KeyError(key)
        return Node(key, (self.basedir or Path()) / key, base=base)

    def __iter__(self):
        """
        Iterate over tuples of path and value.
        """
        for k, _ in self._tree.items():
            yield k

    def items(self):
        """
        Iterate over tuples of ``str`` and  `~dirlay.Node` objects relative to
        layout root.
        """
        for k in self._tree.keys():
            parent, _ = self._tree.traverse(k)
            yield k, Node(k, (self.basedir or Path()) / k, base=parent)

    def keys(self):
        """
        Get all string paths relative to layout root.
        """
        return self._tree.keys()

    def values(self):
        """
        Get all `~dirlay.Node` objects relative to layout root.
        """
        return tuple(v for _, v in self.items())

    def root(self):
        """
        Get root `~dirlay.Node` object.
        """
        return Node('.', self.basedir or Path(), {'.': self._tree.data})

    def leaves(self):
        """
        Get all `~dirlay.Node` objects representing files or empty directories.
        """
        return tuple(n for n in self.values() if not n.is_dir or n.data == {})

    def __or__(self, entries):
        """
        Append dict of entries to a copy of self.

        Equivalent to ``x = self.copy(); x.update(entries)``.
        """
        ret = self.copy()
        ret.update(entries)
        return ret

    def __ior__(self, entries):
        """
        Append dict of entries to self.

        Equivalent to ``self.update(entries)``.
        """
        self.update(entries)
        return self

    def __enter__(self):
        """
        Enter context manager.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit context manager.
        """
        if self.basedir is not None:
            self.rmtree()

    def update(self, entries):
        """
        Update or add entries from dictionary.
        """
        self._tree.update(entries)

    def copy(self):
        """
        Return a deep copy of self.
        """
        return Dir(self._tree.data)

    # filesystem operations

    @property
    def basedir(self):
        """
        Base filesystem directory as `~pathlib.Path` object.

        When ``None``, directory layout object is not instantiated (not created on the
        file system).
        """
        return None if self._basedir is None else self._basedir

    def mktree(self, basedir=None, chdir=None):
        """
        Create directories and files in given or temporary directory.

        Args:

            basedir (`~pathlib.Path` | ``str`` | ``None``, optional):
                Path to base directory under which directories and files will be
                created; if ``None`` (default), temporary directory is used. After the
                directory structure is created, ``basedir`` value is available as
                `~dirlay.Dir.basedir` attribute.

            chdir (`~pathlib.Path` | ``str`` | ``bool`` | ``None``, optional):
                Change the current directory to given path. If ``None`` (default) or
                ``False``, directory is not changed; ``True`` is equivalent to ``'.'``.

        Returns:

            ``None``

        Raises:

            FileExistsError: If ``basedir`` path already exists.
        """
        # prepare
        if basedir is None:
            self._basedir = Path(mkdtemp())
            self._basedir_remove = True
        else:
            basedir = Path(basedir)
            if not basedir.exists():
                basedir.mkdir(parents=True, exist_ok=True)
                self._basedir_remove = True
            self._basedir = basedir.resolve()
        # create
        for node in self.leaves():
            path = self._basedir / node.path
            if node.is_dir:
                path.mkdir(parents=True, exist_ok=True)
            else:
                path.parent.mkdir(parents=True, exist_ok=True)
                if sys.version_info > (3,):
                    path.write_text(node.data)
                else:  # pragma: no cover
                    path.write_text(node.data.decode('utf-8'))
        # chdir
        if chdir not in (None, False):
            self.chdir('.' if chdir is True else chdir)
        #
        return self

    def rmtree(self):
        """
        Remove directory and all its contents.

        If ``basedir`` was created, it will be removed. If ``chdir`` argument
        was passed, current working directory will be restored to the original one.

        Returns:

            ``None``
        """
        self._assert_tree_created()
        # chdir back if needed
        if self._original_cwd is not None:
            os.chdir(str(self._original_cwd))
            self._original_cwd = None
        # remove basedir if needed
        if self._basedir_remove:
            if self._basedir.exists():
                shutil.rmtree(str(self._basedir))
            self._basedir_remove = False
        self._basedir = None

    # current directory operations

    def chdir(self, path=None):
        """
        Change current directory to a subdirectory relative to layout base.

        Args:

            path (`~pathlib.Path` | ``str`` | ``None``):
                Relative path to subdirectory to be chdir'ed to; if ``None`` (default),
                `~dirlay.Dir.basedir` will be used.

        Returns:

            ``None``

        Raises:

            ValueError: If ``path`` is absolute.
        """
        # validate type
        self._assert_tree_created()
        if path is None:
            path = Path()
        elif isinstance(path, Path) or isinstance(path, str):
            path = Path(path)
        else:
            raise TypeError('Path must be str or Path object')
        # assert relative
        if path.is_absolute():
            raise ValueError('Absolute path not allowed: "{}"'.format(path))
        # chdir
        if self._original_cwd is None:
            self._original_cwd = getcwd()
        os.chdir(str(self.basedir / path))

    def _assert_tree_created(self):
        if self._basedir is None:
            raise RuntimeError('Directory tree must be created')

    # formatting

    def as_rich(self, real_basedir=False, show_data=False, **kwargs):
        """
        Return :external+rich:py:obj:`~rich.tree.Tree` representation;
        `rich <https://rich.readthedocs.io>`_ must be installed.
        See :ref:`Print as tree` for examples.

        Args:

            real_basedir (``bool``):
                Whether to show real base directory name instead of ``'.'``; defaults to
                ``False``.

            show_data (``bool``):
                Whether to include file content in the box under the file name; defaults
                to ``False``.

            kwargs (``Any``):
                Optional keyword arguments passed to `~rich.tree.Tree`.

        Returns:

            :external+rich:py:obj:`~rich.tree.Tree`
        """
        if rich is None:
            raise NotImplementedError('Optional dependency required: dirlay[rich]')
        return as_tree(self, real_basedir=real_basedir, show_data=show_data, **kwargs)

    def print_rich(self, real_basedir=False, show_data=False, **kwargs):
        """
        Print :external+rich:py:obj:`~rich.tree.Tree` representation.
        See `~dirlay.Dir.as_rich`.

        Returns:

            ``None``
        """
        if rich is None:
            raise NotImplementedError('Optional dependency required: dirlay[rich]')
        tree = self.as_rich(real_basedir=real_basedir, show_data=show_data, **kwargs)
        rich_print(tree)


# public helpers


def getcwd():
    """
    Get current working directory.

    Works for Python 2 and 3.

    Returns:

        `~pathlib.Path`
    """
    return Path.cwd().resolve()


# internal helpers


def norm(path):
    return os.path.normpath(str(path))



from unittest import TestCase

from dirlay.nested_dict import NestedDict


class TestNestedDict(TestCase):
    # __init__()

    def test_init_empty(self):  # type: () -> None
        self.assertEqual({}, NestedDict())
        self.assertEqual({}, NestedDict(None))
        self.assertEqual({}, NestedDict({}))

    def test_init_invalid(self):  # type: () -> None
        with self.assertRaises(TypeError):
            NestedDict('invalid')  # type: ignore

    def test_init_nested(self):  # type: () -> None
        self.assertEqual({'a': {'b': 'c'}}, NestedDict({'a': {'b': 'c'}}))
        self.assertEqual({'a': {'b': 'c'}}, NestedDict({'a/b': 'c'}))
        self.assertEqual(
            {'a': {'b': 'c', 'd': 'e', 'f': {'g': 'h'}}},
            NestedDict({'a/b': 'c', 'a/d': 'e', 'a/f/g': 'h'}),
        )

    # __eq__()

    # note: positive __eq__() cases are tested all around

    def test_not_equal(self):  # type: () -> None
        self.assertNotEqual({'a': 'b'}, NestedDict({'a': 'c'}))
        self.assertNotEqual((), NestedDict({'a': 'c'}))

    # __len__()

    def test_len(self):  # type: () -> None
        self.assertEqual(0, len(NestedDict({})))
        self.assertEqual(1, len(NestedDict({'a': 'b'})))
        self.assertEqual(2, len(NestedDict({'a': {'b': 'c'}})))
        self.assertEqual(5, len(NestedDict({'a/b/c/d/e': 'f'})))
        self.assertEqual(4, len(NestedDict({'a/b/c': 'd', 'a/b/e': 'f'})))

    # __getitem__()

    def test_getitem(self):  # type: () -> None
        self.assertEqual('b', NestedDict({'a': 'b'})['a'])
        self.assertEqual({'c': 'd'}, NestedDict({'a/b/c': 'd'})['a/b'])

    def test_getitem_error(self):  # type: () -> None
        with self.assertRaises(KeyError):
            _ = NestedDict()['missing']

    # __setitem__()

    def test_setitem(self):  # type: () -> None
        for src, item, expected in (
            ({}, ('a', 'b'), {'a': 'b'}),
            ({'x': 'y'}, ('a/b', {'c': 'd'}), {'a': {'b': {'c': 'd'}}, 'x': 'y'}),
        ):
            result = NestedDict(src)  # type: ignore
            result[item[0]] = item[1]
            self.assertEqual(expected, result)

    def test_setitem_error(self):  # type: () -> None
        with self.assertRaises(KeyError):
            _ = NestedDict()['missing']

    # __delitem__()

    def test_delitem(self):  # type: () -> None
        for src, key, expected in (
            ({'a': 'b'}, 'a', {}),
            ({'a': {'b': {'c': 'd'}}, 'x': 'y'}, 'a/b', {'a': {}, 'x': 'y'}),
            ({'a': {'b': {'c': 'd'}}, 'x': 'y'}, 'a', {'x': 'y'}),
        ):
            result = NestedDict(src)  # type: ignore
            del result[key]
            self.assertEqual(expected, result)

    def test_delitem_error(self):  # type: () -> None
        with self.assertRaises(KeyError):
            del NestedDict()['missing']

    # __contains__()

    def test_contains(self):  # type: () -> None
        for src, key, expected in (
            ({}, 'whatever', False),
            ({'a': 'b'}, 'a', True),
            ({'a': {'b': {'c': 'd'}}}, 'a/b/c', True),
            # invalid use
            ({'a': 'b'}, 0, False),
            ({'a': 'b'}, {}, False),
        ):
            self.assertEqual(expected, key in NestedDict(src))

    # update()

    def test_update(self):  # type: () -> None
        for src, upd, expected in (
            ({}, {'a': 'b'}, {'a': 'b'}),
            ({'a/b': 'c'}, {'a/d': 'e'}, {'a': {'b': 'c', 'd': 'e'}}),
            ({'a/b': 'c'}, {'a': {'d': 'e'}}, {'a': {'b': 'c', 'd': 'e'}}),
        ):
            self.assertEqual(expected, (NestedDict(src) | upd).data)



try:
    from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
    TYPE_CHECKING = False

if TYPE_CHECKING:  # pragma: no cover
    from typing import Optional  # noqa: F401  # used in type hint comment

from rich import print as rich_print  # noqa: F401  # rich_print is exported
from rich.box import ROUNDED
from rich.panel import Panel
from rich.tree import Tree

try:
    from rich.console import Group
except ImportError:  # pragma: no cover
    from rich.group import Group

from .optional import pathlib


Path = pathlib.Path


class DefaultTheme:
    style = 'tree'  # type: str
    guide_style = 'tree.line'  # type: str
    icon_dir = ':open_file_folder:'  # type: str
    icon_file = ':page_facing_up:'  # type: str
    content_box = ROUNDED


def as_tree(layout, real_basedir=False, show_data=False, **kwargs):
    """
    Return :external+rich:py:obj:`~rich.tree.Tree` object representing
    the directory layout. See :ref:`Use cases` for examples.

    Args:

        layout (`~dirlay.Dir`):
            Directory layout to be formatted.

        real_basedir (``bool``):
            Whether to show real base directory name instead of ``'.'``; defaults to
            ``False``.

        show_data (``bool``):
            Whether to include file content in the box under the file name; defaults to
            ``False``.

        kwargs (``Any``):
            Optional keyword arguments passed to `~rich.tree.Tree`.

    Returns:

        ``None``
    """
    theme = DefaultTheme

    def label(item, real_path=False):  # type: (Node, bool) -> str
        icon_type = theme.icon_dir if item.is_dir else theme.icon_file
        icon = '' if icon_type is None else '{} '.format(icon_type)
        filename = item.path if real_path else (item.path.name or '.')
        return '{}{}'.format(icon, filename)

    root = layout.root()
    root.path = Path((layout.basedir or '.') if real_basedir else '.')
    tree = Tree(label(root, real_path=real_basedir), **kwargs)
    nodes = {'.': tree}

    for item in sorted(list(layout.values()), key=lambda item: item.key):
        node = (
            Group(label(item), Panel(item.data, theme.content_box, expand=False))
            if show_data and not item.is_dir
            else label(item)
        )
        base = nodes[str(Path(item.key).parent)]
        nodes[item.key] = base.add(node)

    return tree



# encoding: utf-8
from doctest import ELLIPSIS
import sys
from unittest import TestCase, skipIf

from doctestcase import doctestcase

from dirlay import Dir, Path, getcwd


case = doctestcase(
    globals={'Dir': Dir, 'Path': Path, 'getcwd': getcwd},
    options=ELLIPSIS,
)


@case
class QuickStart(TestCase):
    """
    QuickStart

    Define directory structure and files content:

    >>> layout = Dir({'a': {'b/c.txt': 'ccc', 'd.txt': 'ddd'}})
    >>> layout.data == {'a': {'b': {'c.txt': 'ccc'}, 'd.txt': 'ddd'}}
    True
    >>> layout['a/b/c.txt']
    <Node 'a/b/c.txt': 'ccc'>
    >>> 'z.txt' in layout
    False

    Content of files and directories can be updated:

    >>> layout |= {'a/d.txt': {'e.txt': 'eee'}}
    >>> layout['a/b/c.txt'].data *= 2
    >>> layout.root()
    <Node '.': {'a': {...}}>
    >>> layout.data == {'a': {'b': {'c.txt': 'cccccc'}, 'd.txt': {'e.txt': 'eee'}}}
    True

    Instantiate on the file system (in temporary directory by default) and remove when
    exiting the context.

    >>> with layout.mktree():
    ...     assert getcwd() != layout.basedir  # cwd not changed
    ...     str(layout['a/b/c.txt'].path.read_text())
    'cccccc'

    Optionally, change current working directory to a layout subdir, and change back
    after context manager is exited.

    >>> with layout.mktree(chdir='a/b'):
    ...     assert getcwd() == layout.basedir / 'a/b'
    ...     str(Path('c.txt').read_text())
    'cccccc'
    """


@case
class UsageCreate(TestCase):
    """
    Create directory layout tree

    Directory layout can be constructed from dict:

    >>> layout = Dir({'a': {'b/c.txt': 'ccc', 'd.txt': 'ddd'}})
    >>> layout.basedir is None
    True
    >>> layout.mktree()
    <Dir '/tmp/...': {'a': ...}>
    >>> layout.basedir
    PosixPath('/tmp/...')

    And remove when not needed anymore:

    >>> layout.rmtree()
    """


@case
class UsageChdir(TestCase):
    """
    Chdir to subdirectory

    >>> import os
    >>> os.chdir('/tmp')

    When layout is instantiated, current directory remains unchanged:

    >>> layout = Dir({'a/b/c.txt': 'ccc'})
    >>> layout.mktree()
    <Dir '/tmp/...': {'a': {'b': {'c.txt': 'ccc'}}}>
    >>> getcwd()
    PosixPath('/tmp')

    On first `chdir`, initial working directory is stored internally, and will be
    restored on `destroy`. Without argument, `chdir` sets current directory to
    `layout.basedir`.

    >>> layout.basedir
    PosixPath('/tmp/...')
    >>> layout.chdir()
    >>> getcwd()
    PosixPath('/tmp/...')

    If `chdir` has argument, it must be a path relative to `basedir`.

    >>> layout.chdir('a/b')
    >>> getcwd()
    PosixPath('/tmp/.../a/b')

    When directory is removed, current directory is restored:

    >>> layout.rmtree()
    >>> getcwd()
    PosixPath('/tmp')
    """


@skipIf(sys.version_info < (3, 6), 'rich not supported')
@case
class UsageTree(TestCase):
    """
    Print as tree

    >>> layout = Dir({'a': {'b/c.txt': 'ccc', 'd.txt': 'ddd'}})
    >>> layout.print_rich()
    📂 .
    └── 📂 a
        ├── 📂 b
        │   └── 📄 c.txt
        └── 📄 d.txt

    Display `basedir` path and file content:

    >>> layout.mktree()
    <Dir '/tmp/...': ...>
    >>> layout.print_rich(real_basedir=True, show_data=True)
    📂 /tmp/...
    └── 📂 a
        ├── 📂 b
        │   └── 📄 c.txt
        │       ╭─────╮
        │       │ ccc │
        │       ╰─────╯
        └── 📄 d.txt
            ╭─────╮
            │ ddd │
            ╰─────╯

    Extra keyword arguments will be passed through to `rich.tree.Tree`:

    >>> layout.print_rich(show_data=True, hide_root=True)
    📂 a
    ├── 📂 b
    │   └── 📄 c.txt
    │       ╭─────╮
    │       │ ccc │
    │       ╰─────╯
    └── 📄 d.txt
        ╭─────╮
        │ ddd │
        ╰─────╯

    >>> layout.rmtree()
    """



import re
from unittest import TestCase

from caseutil import to_kebab
from docsub import click
from doctestcase import get_title, to_markdown
from importloc import Location, get_subclasses, random_name


@click.group()
def x() -> None:
    pass


@x.command()
@click.argument('testpath', type=click.Path(exists=True, dir_okay=False))
@click.argument('nameregex')
@click.option('--title/--no-title', default=True)
def cases(testpath: click.Path, nameregex: str, title: bool) -> None:
    """
    Print usage section based on test case docstrings.
    """
    cases = get_subclasses(Location(str(testpath)).load(random_name), TestCase)
    cases.sort(key=lambda c: c.__firstlineno__)  # type: ignore[attr-defined]
    for case in cases:
        if re.fullmatch(nameregex, case.__name__):
            click.echo(to_markdown(case, include_title=title))


@x.command()
@click.argument('testpath', type=click.Path(exists=True, dir_okay=False))
@click.argument('nameregex')
def toc(testpath: click.Path, nameregex: str) -> None:
    """
    Print ToC based on test case docstrings.
    """
    cases = get_subclasses(Location(str(testpath)).load(random_name), TestCase)
    cases.sort(key=lambda c: c.__firstlineno__)  # type: ignore[attr-defined]
    for case in cases:
        if re.fullmatch(nameregex, case.__name__):
            title = get_title(case)
            click.echo(f'* [{title}](#{to_kebab(title)})')



from collections.abc import Iterable
from typing import Any, MutableMapping, Optional, Tuple, Union

from typing_extensions import TypeAlias

from dirlay.nested_dict import NestedDict
from dirlay.optional import rich
from dirlay.types import DictTree, DictNode, Path as Path, PathType

if rich is not None:
    from rich.tree import Tree as RichTree
else:
    RichTree = None  # type: ignore[misc]  # assign to type

class Node(object):
    key: str
    path: Path
    is_dir: bool
    data: 'DictNode'
    _base: DictTree
    def __init__(self, key: str, path: PathType, base: DictTree) -> None: ...
    def __eq__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...

MutableDictTree: TypeAlias = MutableMapping[str, 'MutableDictNode']
MutableDictNode: TypeAlias = Union[MutableDictTree, str]

class Dir:
    _tree: NestedDict[MutableDictTree]
    _basedir: Optional[Path]
    _basedir_remove: bool
    _original_cwd: Optional[Path]
    def __init__(self, entries: Optional[DictTree] = ...) -> None: ...
    @property
    def data(self) -> DictTree: ...
    def __contains__(self, path: PathType) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __getitem__(self, path: PathType) -> Node: ...
    def __iter__(self) -> Iterable[str]: ...
    def items(self) -> Iterable[Tuple[str, Node]]: ...
    def keys(self) -> Tuple[str]: ...
    def values(self) -> Tuple[Node]: ...
    def root(self) -> Node: ...
    def leaves(self) -> Tuple[Node]: ...
    def __or__(self, entries: Union[Dir, DictTree]) -> 'Dir': ...
    def __ior__(self, entries: Union[Dir, DictTree]) -> 'Dir': ...
    def __enter__(self) -> 'Dir': ...
    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> 'Dir': ...
    def copy(self) -> 'Dir': ...
    def update(self, entries: Union[Dir, DictTree], exist_ok: bool = ...) -> None: ...
    def _assert_tree_created(self) -> None: ...
    @property
    def basedir(self) -> Optional[Path]: ...
    def chdir(self, path: Optional[PathType] = ...) -> None: ...
    def mktree(
        self,
        basedir: Optional[PathType] = ...,
        chdir: Union[PathType | bool | None] = ...,
    ) -> 'Dir': ...
    def rmtree(self) -> None: ...
    def as_rich(
        self,
        real_basedir: bool = ...,
        show_data: bool = ...,
        **kwargs: Any,
    ) -> RichTree: ...
    def print_rich(
        self,
        real_basedir: bool = ...,
        show_data: bool = ...,
        **kwargs: Any,
    ) -> None: ...

def getcwd() -> Path: ...



from collections import UserDict
from collections.abc import ItemsView, Iterable, Iterator, KeysView, ValuesView
from typing import Any, Generic, MutableMapping, Optional, Tuple, Type, TypeVar

from dirlay.types import AnyDict, StrDict

D = TypeVar('D', bound=MutableMapping[str, Any])

class NestedDict(UserDict[str, Any], Generic[D]):
    dict_class: Type[D]
    data: D  # type: ignore[assignment]
    sep: str
    _len: int
    def __init__(self, dict: Optional[StrDict] = None, sep: str = ...): ...
    def __eq__(self, other: Any) -> bool: ...
    def __len__(self) -> int: ...
    def __getitem__(self, key: str) -> Any: ...
    def __setitem__(self, key: str, item: Any) -> None: ...
    def __delitem__(self, key: str) -> None: ...
    def _count(self, item: Any) -> int: ...
    def traverse(self, key: str) -> Tuple[StrDict, str]: ...
    def _traverse(
        self,
        key: str,
        create_parents: bool,
        base: StrDict,
    ) -> Tuple[StrDict, str]: ...
    def update(self, other: Optional[StrDict] = ..., **kwargs: Any) -> None: ...  # type: ignore[override]
    def _update(self, other: StrDict, base: StrDict) -> None: ...
    def __iter__(self) -> Iterator[str]: ...
    def items(self) -> ItemsView[str, Any]: ...
    def keys(self) -> KeysView[str]: ...
    def values(self) -> ValuesView[str]: ...
    def _walk(
        self,
        entries: StrDict,
        prefix: Optional[str] = ...,
    ) -> Iterable[Tuple[str, Any, StrDict]]: ...
    def __contains__(self, key: object) -> bool: ...
    def get(self, key: str, default: Any = ...) -> Any: ...
    def __repr__(self) -> str: ...
    def __or__(self, other: AnyDict) -> 'NestedDict': ...  # type: ignore
    def __ror__(self, other: AnyDict) -> 'NestedDict': ...  # type: ignore
    def __ior__(self, other: AnyDict) -> 'NestedDict': ...  # type: ignore
    @staticmethod
    def _operand(other: AnyDict) -> StrDict: ...
    def __copy__(self) -> 'NestedDict[D]': ...
    def clear(self) -> None: ...
    def copy(self) -> 'NestedDict[D]': ...
    @classmethod
    def fromkeys(cls, iterable: Iterable[str], value: Any = ...) -> 'NestedDict[D]': ...  # type: ignore
